@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997-2004  Motoyuki Kasahara
+ * Copyright (c) 1997-2006  Motoyuki Kasahara
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -35,7 +35,7 @@
 /*
  * The maximum number of arguments for an escape sequence.
  */
-#define EB_MAX_ARGV 	6
+#define EB_MAX_ARGV 	7
 
 /*
  * Read next when the length of cached data is shorter than this value.
@@ -213,7 +213,7 @@ eb_seek_text(EB_Book *book, const EB_Position *position)
      */
     eb_reset_text_context(book);
     book->text_context.code = EB_TEXT_SEEKED;
-    book->text_context.location = (position->page - 1) * EB_SIZE_PAGE
+    book->text_context.location = ((off_t) position->page - 1) * EB_SIZE_PAGE
 	+ position->offset;
 
     /*
@@ -335,7 +335,7 @@ eb_read_text(EB_Book *book, EB_Appendix *appendix, EB_Hookset *hookset,
 	    && position.page <= book->subbook_current->copyright.end_page)
 	    book->text_context.code = EB_TEXT_OPTIONAL_TEXT;
 	else
-	    book->text_context.code = EB_TEXT_TEXT;
+	    book->text_context.code = EB_TEXT_MAIN_TEXT;
 
 	hook = hookset->hooks + EB_HOOK_INITIALIZE;
 	if (hook->function != NULL) {
@@ -344,7 +344,7 @@ eb_read_text(EB_Book *book, EB_Appendix *appendix, EB_Hookset *hookset,
 	    if (error_code != EB_SUCCESS)
 		goto failed;
 	}
-    } else if (book->text_context.code != EB_TEXT_TEXT
+    } else if (book->text_context.code != EB_TEXT_MAIN_TEXT
 	&& book->text_context.code != EB_TEXT_OPTIONAL_TEXT) {
 	error_code = EB_ERR_DIFF_CONTENT;
 	goto failed;
@@ -732,7 +732,7 @@ text_max_length=%ld, forward=%d)",
 		argv[1] = eb_uint2(cache_p + 2);
 
 		if (0 < context->printable_count
-		    && context->code == EB_TEXT_TEXT) {
+		    && context->code == EB_TEXT_MAIN_TEXT) {
 		    if (eb_is_stop_code(book, appendix, argv[0], argv[1])) {
 			context->text_status = EB_TEXT_STATUS_SOFT_STOP;
 			goto succeeded;
@@ -863,7 +863,7 @@ text_max_length=%ld, forward=%d)",
 		argv[1] = eb_uint2(cache_p + 2);
 
 		if (0 < context->printable_count
-		    && context->code == EB_TEXT_TEXT) {
+		    && context->code == EB_TEXT_MAIN_TEXT) {
 		    if (eb_is_stop_code(book, appendix, argv[0], argv[1])) {
 			context->text_status = EB_TEXT_STATUS_SOFT_STOP;
 			goto succeeded;
@@ -956,6 +956,7 @@ text_max_length=%ld, forward=%d)",
 		argv[1] = eb_bcd4(cache_p + 2);
 		argv[2] = eb_bcd2(cache_p + 6);
 		if (cache_p[8]==0x1f && cache_p[9]==0x6b) {
+		    context->text_status = EB_TEXT_STATUS_SOFT_STOP;
 		    hook = hookset->hooks + EB_HOOK_GRAPHIC_REFERENCE;
 		    in_step = 10;
 		} else {
@@ -963,6 +964,16 @@ text_max_length=%ld, forward=%d)",
 		}
 		break;
 
+	    case 0x4c:
+		/* beginning of image page */
+		in_step = 4;
+		if (cache_rest_length < in_step) {
+		    error_code = EB_ERR_UNEXP_TEXT;
+		    goto failed;
+		}
+		hook = hookset->hooks + EB_HOOK_BEGIN_IMAGE_PAGE;
+		break;
+
 	    case 0x4d:
 		/* beginning of color graphic (BMP or JPEG) */
 		in_step = 20;
@@ -980,7 +991,25 @@ text_max_length=%ld, forward=%d)",
 		    hook = hookset->hooks + EB_HOOK_BEGIN_COLOR_JPEG;
 		break;
 
-	    case 0x49: case 0x4c: case 0x4e: case 0x4f:
+	    case 0x4f:
+	        /* beginning of clickable area */
+	        in_step = 34;
+	        if (cache_rest_length < in_step) {
+		    error_code = EB_ERR_UNEXP_TEXT;
+		    goto failed;
+		}
+		argc = 7;
+		argv[1] = eb_bcd2(cache_p + 8);	  
+		argv[2] = eb_bcd2(cache_p + 10);	  
+		argv[3] = eb_bcd2(cache_p + 12);	  
+		argv[4] = eb_bcd2(cache_p + 14);	  
+		argv[5] = eb_bcd4(cache_p + 28);	  
+		argv[6] = eb_bcd2(cache_p + 32);	  
+		hook = hookset->hooks + EB_HOOK_BEGIN_CLICKABLE_AREA;
+	      break;
+
+	    case 0x49: case 0x4e:
+
 		in_step = 2;
 		context->skip_code = eb_uint1(cache_p + 1) + 0x20;
 		break;
@@ -1095,6 +1124,17 @@ text_max_length=%ld, forward=%d)",
 		hook = hookset->hooks + EB_HOOK_END_WAVE;
 		break;
 
+	    case 0x6c:
+		/* end of image page */
+		in_step = 2;
+		if (cache_rest_length < in_step) {
+		    error_code = EB_ERR_UNEXP_TEXT;
+		    goto failed;
+		}
+		context->text_status = EB_TEXT_STATUS_SOFT_STOP;
+		hook = hookset->hooks + EB_HOOK_END_IMAGE_PAGE;
+		break;
+
 	    case 0x6d:
 		/* end of color graphic (BMP or JPEG) */
 		in_step = 2;
@@ -1105,6 +1145,16 @@ text_max_length=%ld, forward=%d)",
 		hook = hookset->hooks + EB_HOOK_END_COLOR_GRAPHIC;
 		break;
 
+	    case 0x6f:
+		/* end of clickable area */
+		in_step = 2;
+		if (cache_rest_length < in_step) {
+		    error_code = EB_ERR_UNEXP_TEXT;
+		    goto failed;
+		}
+		hook = hookset->hooks + EB_HOOK_END_CLICKABLE_AREA;
+		break;
+
 	    case 0x70: case 0x71: case 0x72: case 0x73: case 0x74: case 0x75:
 	    case 0x76: case 0x77: case 0x78: case 0x79: case 0x7a: case 0x7b:
 	    case 0x7c: case 0x7d: case 0x7e: case 0x7f:
@@ -1372,6 +1422,11 @@ text_max_length=%ld, forward=%d)",
 	 */
 	if (context->unprocessed != NULL)
 	    break;
+	/*
+	 * Break if EB_TEXT_STATUS_SOFT_STOP is set.
+	 */
+	if (context->text_status == EB_TEXT_STATUS_SOFT_STOP)
+	  break;
     }
 
   succeeded:
@@ -1440,7 +1495,7 @@ eb_is_text_stopped(EB_Book *book)
 
     if (book->subbook_current != NULL) {
 	if (book->text_context.code == EB_TEXT_HEADING
-	    || book->text_context.code == EB_TEXT_TEXT
+	    || book->text_context.code == EB_TEXT_MAIN_TEXT
 	    || book->text_context.code == EB_TEXT_OPTIONAL_TEXT) {
 	    if (book->text_context.text_status != EB_TEXT_STATUS_CONTINUED
 		&& book->text_context.unprocessed == NULL) {
@@ -1695,11 +1750,11 @@ eb_forward_text(EB_Book *book, EB_Appendix *appendix)
     }
 
     if (book->text_context.code == EB_TEXT_SEEKED) {
-	book->text_context.code = EB_TEXT_TEXT;
+	book->text_context.code = EB_TEXT_MAIN_TEXT;
     } else if (book->text_context.code == EB_TEXT_INVALID) {
 	error_code = EB_ERR_NO_PREV_SEEK;
 	goto failed;
-    } else if (book->text_context.code != EB_TEXT_TEXT
+    } else if (book->text_context.code != EB_TEXT_MAIN_TEXT
 	&& book->text_context.code != EB_TEXT_OPTIONAL_TEXT) {
 	error_code = EB_ERR_DIFF_CONTENT;
 	goto failed;
@@ -1754,6 +1809,25 @@ eb_forward_heading(EB_Book *book)
     eb_lock(&book->lock);
     LOG(("in: eb_forward_heading(book=%d)", (int)book->code));
 
+    if (book->subbook_current == NULL) {
+	error_code = EB_ERR_NO_CUR_SUB;
+	goto failed;
+    }
+    if (zio_file(&book->subbook_current->text_zio) < 0) {
+	error_code = EB_ERR_NO_TEXT;
+	goto failed;
+    }
+
+    if (book->text_context.code == EB_TEXT_SEEKED) {
+	book->text_context.code = EB_TEXT_HEADING;
+    } else if (book->text_context.code == EB_TEXT_INVALID) {
+	error_code = EB_ERR_NO_PREV_SEEK;
+	goto failed;
+    } else if (book->text_context.code != EB_TEXT_HEADING) {
+	error_code = EB_ERR_DIFF_CONTENT;
+	goto failed;
+    }
+
     if (book->text_context.text_status == EB_TEXT_STATUS_SOFT_STOP) {
 	book->text_context.text_status = EB_TEXT_STATUS_CONTINUED;
 	goto succeeded;
@@ -1762,11 +1836,6 @@ eb_forward_heading(EB_Book *book)
 	goto failed;
     }
 
-    /*
-     * Force setting context mode as heading mode.
-     */
-    book->text_context.code = EB_TEXT_HEADING;
-
     /*
      * Forward text.
      */
@@ -1831,11 +1900,11 @@ eb_backward_text(EB_Book *book, EB_Appendix *appendix)
     }
 
     if (book->text_context.code == EB_TEXT_SEEKED) {
-	book->text_context.code = EB_TEXT_TEXT;
+	book->text_context.code = EB_TEXT_MAIN_TEXT;
     } else if (book->text_context.code == EB_TEXT_INVALID) {
 	error_code = EB_ERR_NO_PREV_SEEK;
 	goto failed;
-    } else if (book->text_context.code != EB_TEXT_TEXT
+    } else if (book->text_context.code != EB_TEXT_MAIN_TEXT
 	&& book->text_context.code != EB_TEXT_OPTIONAL_TEXT) {
 	error_code = EB_ERR_DIFF_CONTENT;
 	goto failed;
@@ -1948,7 +2017,7 @@ eb_backward_text(EB_Book *book, EB_Appendix *appendix)
 		backward_location = book->text_context.location;
 		goto loop_end;
 	    }
-	    if (book->text_context.code != EB_TEXT_TEXT
+	    if (book->text_context.code != EB_TEXT_MAIN_TEXT
 		|| eb_uint2(text_buffer_p) != stop_code0
 		|| eb_uint2(text_buffer_p + 2) != stop_code1) {
 		text_buffer_p--;
